//+------------------------------------------------------------------+
//|                                           cMessagePipeServer.mqh |
//|                                                            micle |
//|                            https://login.mql5.com/ru/users/micle |
//+------------------------------------------------------------------+
#property copyright "micle"
#property link      "https://login.mql5.com/ru/users/micle"
#property version   "1.00"

#include "cMessagePipeCommon.mqh"

class cMessagePipeServer{

private:
   int      m_PipeCount;
   int      m_Pipes[];
   string   m_PipeName;
   
   int      CreatePipeInstance();
   string   CheckPipe(int PipeIndex);
   
public:
            cMessagePipeServer();
            cMessagePipeServer(string PipeName, int UsePipeInstances = DEFAULT_MAX_PIPES);
            ~cMessagePipeServer();
   int      CheckForPipeMessages(string & arrMessages[]);
};

cMessagePipeServer::cMessagePipeServer(string PipeName, int UsePipeInstances = DEFAULT_MAX_PIPES)
{
   // Store the number of pipe instances to use and resize the array accordinging
   m_PipeCount = UsePipeInstances;
   ArrayResize(m_Pipes, m_PipeCount);
   // Store the name to use for the pipe instances
   m_PipeName = PipeName;

   // Create the pipe instances
   for (int i = 0; i < m_PipeCount; i++) {
      m_Pipes[i] = CreatePipeInstance();
   }
   return;
}

cMessagePipeServer::~cMessagePipeServer()
{
   for (int i = 0; i < m_PipeCount; i++) {
      int handle = m_Pipes[i];
      CloseHandle(handle);
   }
   return;
}

int cMessagePipeServer::CreatePipeInstance()
{
   string strPipeName = PIPE_BASE_NAME + m_PipeName;
   int openMode = PIPE_ACCESS_DUPLEX;
   int pipeMode = PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE | PIPE_NOWAIT;
   int pipeInstanses = PIPE_UNLIMITED_INSTANCES;
   Print("CreatePipeInstance::call CreateNamedPipeW");
   int handle = CreateNamedPipeW(strPipeName, openMode, pipeMode, pipeInstanses, 1000, 1000, 0, 0);
   Print("CreatePipeInstance::return");
   return(handle);
}

// Checks for new messages. The return value is the number of messages received since the last check,
// and the array is resized to contain the incoming messages
int cMessagePipeServer::CheckForPipeMessages(string & arrMessages[])
{
   int MessagesFound = 0;
   for (int i = 0; i < m_PipeCount; i++) {
      // Check each pipe instance for a message
      string strMsg = CheckPipe(i);
      if (strMsg != "") {

         // If there is a message, add it to the list which we're passing back
         MessagesFound++;
         ArrayResize(arrMessages, MessagesFound);
         arrMessages[MessagesFound - 1] = strMsg;
      }
   }

   return (MessagesFound);
}

// Function which checks to see if a message has come in on a pipe instance. If so, the message
// is retrieved, and the pipe instance is "freed" by destroying and recreating it
string cMessagePipeServer::CheckPipe(int PipeIndex)
{
   string strReturnValue = "";
   // See if there's data available on the pipe
   //int BytesAvailable[1] = {0};
   int BytesAvailable = 0;
   int handle = m_Pipes[PipeIndex];
   int res = PeekNamedPipe(handle, 0, 0, 0, BytesAvailable, 0);
   if (res != 0) {
      // PeekNamedPipe() succeeded
      // Is there data?
      if (BytesAvailable != 0) {
         Print("BytesAvailable=", DoubleToString(BytesAvailable,0));
         // Keep reading until either we have all the data, or an error occurs
         int TotalBytesRead = 0;
         while (TotalBytesRead < BytesAvailable) {
            // Allocate a 200-byte buffer
            string ReadBuffer;
            StringInit(ReadBuffer, 200, 0);
            int BufferLength = StringBufferLen(ReadBuffer);

            // Read up to the maximum buffer size from the pipe. (The use of MulDiv() is an old trick from VB6 days for
            // getting the address in memory of a string variable.)
            int BytesRead = 0;
            //ReadFile(m_Pipes[PipeIndex], ReadBuffer, BufferLength, BytesRead, 0);
            ReadFile(handle, ReadBuffer, BufferLength, BytesRead, 0);
            
            // Did we get any data from the read?
            if (BytesRead > 0) {
               // Yes, got some data. Add it to the total message which is passed back
               strReturnValue = strReturnValue + StringSubstr(ReadBuffer, 0, BytesRead/2);
               TotalBytesRead += BytesRead;
            } else {
               // No, the read failed. Stop reading, and pass back an empty string
               strReturnValue = "";
               break;
            }
         }
         // Destroy and recreate the pipe instance
         CloseHandle(handle);
         m_Pipes[PipeIndex] = CreatePipeInstance();
      } else {
         // No data available on pipe
      }
   } else {
      // PeekNamedPipe() failed
   }

   return (strReturnValue);
}

//+------------------------------------------------------------------+
